package api

import (
	"context"
	"reflect"
	"testing"

	apb "code.fbi.h-da.de/danet/gosdn/api/go/gosdn/rbac"
	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

func TestCreateRoles(t *testing.T) {
	type args struct {
		ctx   context.Context
		addr  string
		roles []*apb.Role
	}
	tests := []struct {
		name    string
		args    args
		want    *apb.CreateRolesResponse
		wantErr bool
	}{
		{
			name: "default create roles",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				roles: []*apb.Role{
					{
						Name:        "new role 1",
						Description: "Role 1",
						Permissions: []string{"permission 1", "permission 2"},
					},
				},
			},
			want:    &apb.CreateRolesResponse{},
			wantErr: false,
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			_, err := CreateRoles(tt.args.ctx, tt.args.addr, tt.args.roles)
			if (err != nil) != tt.wantErr {
				t.Errorf("CreateRoles() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
		})
	}
}

func TestGetRole(t *testing.T) {
	type args struct {
		ctx  context.Context
		addr string
		name string
		id   uuid.UUID
	}
	tests := []struct {
		name    string
		args    args
		want    *apb.GetRoleResponse
		wantErr bool
	}{
		{
			name: "default get role",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				name: "adminTestRole",
				id:   uuid.Nil,
			},
			want: &apb.GetRoleResponse{
				Role: &apb.Role{
					Name:        "adminTestRole",
					Description: "Admin",
				},
			},
			wantErr: false,
		},
		{
			name: "error get role",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				name: "not role",
				id:   uuid.Nil,
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetRole(tt.args.ctx, tt.args.addr, tt.args.name, tt.args.id)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetRole() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			if got != nil {
				if got.GetRole().GetName() != tt.want.GetRole().GetName() || got.GetRole().GetDescription() != tt.want.GetRole().GetDescription() {
					t.Errorf("Role.GetRole() = %v, want %v", got, tt.want)
				}
			} else {
				if got != nil {
					t.Errorf("Role.GetRole() = %v, want %v", got, tt.want)
				}
			}
		})
	}
}

func TestGetRoles(t *testing.T) {
	err := clearAndCreateAuthTestSetup()
	if err != nil {
		t.Fatalf("%v", err)
	}

	type args struct {
		ctx  context.Context
		addr string
	}
	tests := []struct {
		name    string
		args    args
		want    *apb.GetRolesResponse
		wantLen int
		wantErr bool
	}{
		{
			name: "default get roles",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
			},
			want: &apb.GetRolesResponse{
				Roles: []*apb.Role{
					{
						Name:        "adminTestRole",
						Description: "Admin",
						Permissions: []string{
							"/gosdn.core.CoreService/GetPnd",
							"/gosdn.core.CoreService/GetPndList",
						}},
					{
						Name:        "userTestRole",
						Description: "User",
						Permissions: []string{
							"/gosdn.pnd.PndService/GetChangeList",
						}},
					{
						Name:        randomRoleName,
						Description: "Not a role",
						Permissions: []string{
							"nope",
						},
					},
				},
			},
			wantLen: 3,
			wantErr: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetRoles(tt.args.ctx, tt.args.addr)
			if (err != nil) != tt.wantErr {
				t.Errorf("GetRoles() error = %v, wantErr %v", err, tt.wantErr)
				return
			}

			if got != nil {
				if len(got.GetRoles()) != 3 {
					t.Errorf("Role.GetRoles() = %v, want %v", got, tt.want)
				}
				for _, gotR := range got.GetRoles() {
					containsExpected := false
					for _, wantR := range tt.want.GetRoles() {
						gotPerm := gotR.GetPermissions()
						wantPerm := wantR.GetPermissions()
						if gotR.GetDescription() == wantR.GetDescription() && gotR.GetName() == wantR.GetName() &&
							reflect.DeepEqual(gotPerm, wantPerm) {
							containsExpected = true
							break
						}
					}
					if !containsExpected {
						t.Errorf("Role.GetRoles() = %v, want %v", got, tt.want)
					}
				}
			}
		})
	}
}

func TestUpdateRoles(t *testing.T) {
	type args struct {
		ctx   context.Context
		addr  string
		roles []*apb.Role
	}
	tests := []struct {
		name    string
		args    args
		want    *apb.UpdateRolesResponse
		wantErr bool
	}{
		{
			name: "default update roles",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				roles: []*apb.Role{
					{
						Id:          adminRoleID,
						Name:        "New Name",
						Description: "Test",
					},
				},
			},
			want:    &apb.UpdateRolesResponse{},
			wantErr: false,
		},
		{
			name: "error update roles",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				roles: []*apb.Role{
					{
						Id:          uuid.NewString(),
						Name:        "New Name",
						Description: "Test",
					},
				},
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			_, err := UpdateRoles(tt.args.ctx, tt.args.addr, tt.args.roles)
			if (err != nil) != tt.wantErr {
				t.Errorf("UpdateRoles() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
		})
	}
}

func TestDeletePermissionForRole(t *testing.T) {
	if err := clearAndCreateAuthTestSetup(); err != nil {
		log.Error(err)
		t.Fail()
	}

	type args struct {
		ctx                 context.Context
		addr                string
		name                string
		permissionsToDelete []string
	}
	tests := []struct {
		name    string
		args    args
		want    *apb.DeletePermissionsForRoleResponse
		wantErr bool
	}{
		{
			name: "default delete permissions for role",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				name: "adminTestRole",
				permissionsToDelete: []string{
					"/gosdn.core.CoreService/GetPnd",
					"/gosdn.core.CoreService/GetPndList",
				},
			},
			want:    &apb.DeletePermissionsForRoleResponse{},
			wantErr: false,
		},
		{
			name: "error delete permissions for role",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				name: "adminTestRole",
				permissionsToDelete: []string{
					"foo",
				},
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			_, err := DeletePermissionForRole(tt.args.ctx, tt.args.addr, tt.args.name, tt.args.permissionsToDelete)
			if (err != nil) != tt.wantErr {
				t.Errorf("DeletePermissionForRole() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
		})
	}
}

func TestDeleteRoles(t *testing.T) {
	type args struct {
		ctx      context.Context
		addr     string
		roleName []string
	}
	tests := []struct {
		name    string
		args    args
		want    *apb.DeleteRolesResponse
		wantErr bool
	}{
		{
			name: "default delete roles",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				roleName: []string{
					"userTestRole",
					"adminTestRole",
				},
			},
			want:    &apb.DeleteRolesResponse{},
			wantErr: false,
		},
		{
			name: "error delete roles",
			args: args{
				ctx:  context.TODO(),
				addr: testAPIEndpoint,
				roleName: []string{
					"no",
				},
			},
			want:    nil,
			wantErr: true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := clearAndCreateAuthTestSetup(); err != nil {
				log.Error(err)
				t.Fail()
			}

			_, err := DeleteRoles(tt.args.ctx, tt.args.addr, tt.args.roleName)
			if (err != nil) != tt.wantErr {
				t.Errorf("DeleteRoles() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
		})
	}
}
