package server

import (
	"context"
	"time"

	topopb "code.fbi.h-da.de/danet/gosdn/api/go/gosdn/topology"
	query "code.fbi.h-da.de/danet/gosdn/controller/store"
	"code.fbi.h-da.de/danet/gosdn/controller/topology/nodes"
	"code.fbi.h-da.de/danet/gosdn/controller/topology/ports"
	routingtables "code.fbi.h-da.de/danet/gosdn/controller/topology/routing-tables"
	"github.com/bufbuild/protovalidate-go"
	"github.com/google/uuid"

	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"
)

// RoutingTableServiceServer holds a RoutingTableServiceServer and represents a RoutingTableServiceServer.
type RoutingTableServiceServer struct {
	topopb.UnimplementedRoutingTableServiceServer
	routeService   routingtables.Service
	nodeService    nodes.Service
	portService    ports.Service
	protoValidator *protovalidate.Validator
}

// NewRoutingTableServiceServer returns a new RoutingTableServiceServer.
func NewRoutingTableServiceServer(
	service routingtables.Service,
	nodeService nodes.Service,
	portService ports.Service,
	protoValidator *protovalidate.Validator,
) *RoutingTableServiceServer {
	return &RoutingTableServiceServer{
		routeService:   service,
		nodeService:    nodeService,
		portService:    portService,
		protoValidator: protoValidator,
	}
}

// AddRoutingTable adds a new route.
func (t *RoutingTableServiceServer) AddRoutingTable(
	ctx context.Context,
	request *topopb.AddRoutingTableRequest,
) (*topopb.AddRoutingTableResponse, error) {
	if err := t.protoValidator.Validate(request); err != nil {
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}

	routes := []routingtables.Route{}

	for _, incomingRoute := range request.GetRoutingTable().GetRoutes() {
		routes = append(routes, routingtables.Route{
			ID:            getExistingOrCreateNilUUIDFromString(incomingRoute.GetId()),
			TargetIPRange: incomingRoute.GetTargetIPRange(),
			NextHopIP:     incomingRoute.GetNextHopIP(),
			Metric:        incomingRoute.GetMetric(),
			PortID:        getExistingOrCreateNilUUIDFromString(incomingRoute.GetPortID()),
		})
	}

	routeingTable := routingtables.RoutingTable{
		ID:     getExistingOrCreateNilUUIDFromString(request.GetRoutingTable().GetNodeID()),
		NodeID: getExistingOrCreateNilUUIDFromString(request.GetRoutingTable().GetNodeID()),
		Routes: routes,
	}
	_, err := t.routeService.EnsureExists(routeingTable)
	if err != nil {
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}

	return &topopb.AddRoutingTableResponse{
		Timestamp: time.Now().UnixNano(),
	}, nil
}

// GetRoutes returns the current routes.
func (t *RoutingTableServiceServer) GetRoutes(
	ctx context.Context,
	request *topopb.GetRoutesRequest,
) (*topopb.GetRoutesResponse, error) {
	exisitingRoutingTables, err := t.routeService.GetAll()
	if err != nil {
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}

	routingTables := []*topopb.RoutingTable{}

	for _, routingTable := range exisitingRoutingTables {
		rt := topopb.RoutingTable{}

		rt.Id = routingTable.ID.String()
		rt.NodeID = routingTable.NodeID.String()

		for _, ers := range routingTable.Routes {
			rt.Routes = append(rt.Routes, &topopb.Route{
				Id:            ers.ID.String(),
				TargetIPRange: ers.TargetIPRange,
				NextHopIP:     ers.NextHopIP,
				Metric:        ers.Metric,
				PortID:        ers.PortID.String(),
			})
		}

		routingTables = append(routingTables, &rt)
	}

	return &topopb.GetRoutesResponse{
		Timestamp:     time.Now().UnixNano(),
		RoutingTables: routingTables,
	}, nil
}

// DeleteRoute deletes a routing table entry.
func (t *RoutingTableServiceServer) DeleteRoute(
	ctx context.Context,
	request *topopb.DeleteRoutesRequest,
) (*topopb.DeleteRoutesResponse, error) {
	if err := t.protoValidator.Validate(request); err != nil {
		return nil, status.Errorf(codes.Aborted, "%v", err)
	}

	routeID, err := uuid.Parse(request.GetId())
	if err != nil {
		return nil, err
	}

	foundRoute, err := t.routeService.Get(query.Query{ID: routeID})
	if err != nil {
		return nil, err
	}

	err = t.routeService.Delete(foundRoute)
	if err != nil {
		return nil, err
	}

	return &topopb.DeleteRoutesResponse{
		Timestamp: time.Now().UnixNano(),
	}, nil
}
