package integration_test_networkElement

import (
	"context"
	"fmt"
	"os"
	"os/exec"
	"testing"

	mnepb "code.fbi.h-da.de/danet/gosdn/api/go/gosdn/networkelement"
	tpb "code.fbi.h-da.de/danet/gosdn/api/go/gosdn/transport"
	integration_test_utils "code.fbi.h-da.de/danet/gosdn/integration-tests/integrationTestUtils"
	"github.com/openconfig/gnmi/proto/gnmi"
	"github.com/openconfig/ygot/ygot"
	"github.com/stretchr/testify/assert"
	"google.golang.org/grpc"
)

const targetAAdress = "gnmi-target_A:7030"
const targetBAdress = "gnmi-target_B:7030"
const targetUserAndPW = "admin"

var targetA = "localhost:3919"

const pndID = "5f20f34b-cbd0-4511-9ddc-c50cf6a3b49d"
const pluginID = "d1c269a2-6482-4010-b0d8-679dff73153b"
const mneID = "9ee7bf15-15ff-44b8-aafa-035487d1e97f"

// The connection to the controller to use in each test.
var conn *grpc.ClientConn

// The context containing the credentials when authenticated.
var ctx context.Context

// A defaultSDN config with default/empty values.
var defaultSDNConfig string

func TestMain(m *testing.M) {
	localConn, localCtx, err := integration_test_utils.CreateSecureConnection()
	if err != nil {
		fmt.Println(err.Error())
	}
	conn = localConn
	ctx = localCtx

	sndConfig, err := integration_test_utils.ExportCurrentSDNConfig(conn, ctx)
	defaultSDNConfig = sndConfig
	if err != nil {
		fmt.Println(err.Error())
	}

	integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	envVarTargetA := os.Getenv("INTEGRATION_TEST_TARGET_A")
	if envVarTargetA != "" {
		targetA = envVarTargetA
	}

	m.Run()
}

func TestChangeHostnameViaController(t *testing.T) {
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	// setup required parameters
	const initialName = "testTarget"
	const newHostname = "\"newName\"" // json formatted to be sent as typedValue
	const operationUpdate = mnepb.ApiOperation_API_OPERATION_UPDATE
	const expectedHostnameAfterChange = "newName"

	opt := &tpb.TransportOption{
		Address:  targetAAdress,
		Username: targetUserAndPW,
		Password: targetUserAndPW,
		TransportOption: &tpb.TransportOption_GnmiTransportOption{
			GnmiTransportOption: &tpb.GnmiTransportOption{},
		},
		Tls: true,
	}

	addListRequest := &mnepb.AddListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Mne: []*mnepb.SetMne{
			{
				Address:         targetAAdress,
				Pid:             pndID,
				PluginId:        pluginID,
				MneName:         initialName,
				TransportOption: opt,
			},
		},
		Pid: pndID,
	}

	pathToChange, err := ygot.StringToStructuredPath("system/config/hostname")
	if err != nil {
		t.Error(err)
	}

	setPathListRequest := &mnepb.SetPathListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		ChangeRequest: []*mnepb.ChangeRequest{
			{
				Path: pathToChange,
				Value: &gnmi.TypedValue{
					Value: &gnmi.TypedValue_JsonIetfVal{
						JsonIetfVal: []byte(newHostname),
					},
				},
				ApiOp: operationUpdate,
			},
		},
	}

	changeCommit := &mnepb.SetChangeListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Change: []*mnepb.SetChange{
			{
				Op: mnepb.Operation_OPERATION_COMMIT,
			},
		},
		Pid: pndID,
	}

	changeConfirm := &mnepb.SetChangeListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Change: []*mnepb.SetChange{
			{
				Op: mnepb.Operation_OPERATION_CONFIRM,
			},
		},
		Pid: pndID,
	}

	getAllMneRequest := &mnepb.GetAllRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Pid:       pndID,
	}

	getMneRequest := &mnepb.GetRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Pid:       pndID,
	}

	// setup gRPC services
	mneService := mnepb.NewNetworkElementServiceClient(conn)

	// add one device to the controller
	addListResponse, err := mneService.AddList(ctx, addListRequest)
	if err != nil {
		t.Error(err)
	}

	// change hostname
	setPathListRequest.ChangeRequest[0].Mneid = addListResponse.GetResponses()[0].GetId()

	setPathListResponse, err := mneService.SetPathList(ctx, setPathListRequest)
	if err != nil {
		t.Error(err)
	}

	changeCommit.Change[0].Cuid = setPathListResponse.GetResponses()[0].GetId()  // set change ID
	changeConfirm.Change[0].Cuid = setPathListResponse.GetResponses()[0].GetId() // set change ID

	_, err = mneService.SetChangeList(ctx, changeCommit)
	if err != nil {
		t.Error(err)
	}

	_, err = mneService.SetChangeList(ctx, changeConfirm)
	if err != nil {
		t.Error(err)
	}

	// get device to check hostname
	getAllResponse, err := mneService.GetAll(ctx, getAllMneRequest)
	if err != nil {
		t.Error(err)
	}

	getMneRequest.Mneid = getAllResponse.GetMne()[0].GetId()
	getMneResponse, err := mneService.Get(ctx, getMneRequest)
	if err != nil {
		t.Error(err)
	}

	mne := NewMne()

	path, err := ygot.StringToPath("/", ygot.StructuredPath)
	if err != nil {
		t.Error(err)
	}

	err = Unmarshal([]byte(getMneResponse.GetMne().GetModel()), path, mne.Model)
	if err != nil {
		t.Error(err)
		t.FailNow()
	}

	assert.Equal(t, expectedHostnameAfterChange, *mne.Model.System.Config.Hostname)
}

func TestAddMultipleAndDeleteMne(t *testing.T) {
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	// setup required parameters
	const expectedAmountBeforeDeletion = 2
	const expectedAmountAfterDeletion = 1
	const expectedNameForRemainingTarget = "targetA"

	opt := &tpb.TransportOption{
		Address:  targetAAdress,
		Username: targetUserAndPW,
		Password: targetUserAndPW,
		TransportOption: &tpb.TransportOption_GnmiTransportOption{
			GnmiTransportOption: &tpb.GnmiTransportOption{},
		},
		Tls: true,
	}

	addListRequest := &mnepb.AddListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Mne: []*mnepb.SetMne{
			{
				Address:         targetAAdress,
				Pid:             pndID,
				PluginId:        pluginID,
				MneName:         expectedNameForRemainingTarget,
				TransportOption: opt,
				MneId:           mneID,
			},
		},
		Pid: pndID,
	}

	opt.Address = targetBAdress
	addListRequest.Mne = append(addListRequest.Mne,
		&mnepb.SetMne{
			Address:         targetBAdress,
			Pid:             pndID,
			PluginId:        pluginID,
			TransportOption: opt,
		})

	getAllMneRequest := &mnepb.GetAllRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Pid:       pndID,
	}

	deleteSecondTargetRequest := &mnepb.DeleteRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Pid:       pndID,
	}

	// setup gRPC services
	mneService := mnepb.NewNetworkElementServiceClient(conn)

	// add two devices to the controller
	_, err := mneService.AddList(ctx, addListRequest)
	if err != nil {
		t.Error(err)
	}

	// get devices to check amount
	getAllResponse, err := mneService.GetAll(ctx, getAllMneRequest)
	if err != nil {
		t.Error(err)
	}

	assert.Equal(t, expectedAmountBeforeDeletion, len(getAllResponse.GetMne()))

	// delete second device
	deleteSecondTargetRequest.Mneid = getAllResponse.GetMne()[1].GetId()
	_, err = mneService.Delete(ctx, deleteSecondTargetRequest)
	if err != nil {
		t.Error(err)
	}

	// get all devices and compare amount and name
	getAllResponse, err = mneService.GetAll(ctx, getAllMneRequest)
	if err != nil {
		t.Error(err)
	}

	assert.Equal(t, expectedAmountAfterDeletion, len(getAllResponse.GetMne()))

	assert.Equal(t, expectedNameForRemainingTarget, getAllResponse.GetMne()[0].GetName())
}

// Note: This testcase is also covered by TestChangeHostname at the moment.
// Need to discuss if having this as an extra test is necessary, maybe
// more useful if we can get MNE by Name in the future and then also cover this here.
func TestGetMneById(t *testing.T) {
	integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	// setup required parameters
	const mneName = "targetA"

	opt := &tpb.TransportOption{
		Address:  targetAAdress,
		Username: targetUserAndPW,
		Password: targetUserAndPW,
		TransportOption: &tpb.TransportOption_GnmiTransportOption{
			GnmiTransportOption: &tpb.GnmiTransportOption{},
		},
		Tls: true,
	}

	addListRequest := &mnepb.AddListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Mne: []*mnepb.SetMne{
			{
				Address:         targetAAdress,
				Pid:             pndID,
				PluginId:        pluginID,
				MneName:         mneName,
				TransportOption: opt,
			},
		},
		Pid: pndID,
	}

	getByIdRequest := &mnepb.GetRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Pid:       pndID,
	}

	// setup gRPC services
	mneService := mnepb.NewNetworkElementServiceClient(conn)

	// add one device to the controller
	addListResponse, err := mneService.AddList(ctx, addListRequest)
	if err != nil {
		t.Error(err)
	}

	// get by ID and compare
	getByIdRequest.Mneid = addListResponse.GetResponses()[0].GetId()
	getByIdResponse, err := mneService.Get(ctx, getByIdRequest)
	if err != nil {
		t.Error(err)
	}

	assert.Equal(t, mneName, getByIdResponse.GetMne().GetName())
}

func TestGetRootPathAndPushToMne(t *testing.T) {
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	// setup required parameters
	const mneName = "targetA"
	const operationUpdate = mnepb.ApiOperation_API_OPERATION_UPDATE

	opt := &tpb.TransportOption{
		Address:  targetAAdress,
		Username: targetUserAndPW,
		Password: targetUserAndPW,
		TransportOption: &tpb.TransportOption_GnmiTransportOption{
			GnmiTransportOption: &tpb.GnmiTransportOption{},
		},
		Tls: true,
	}

	addListRequest := &mnepb.AddListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Mne: []*mnepb.SetMne{
			{
				Address:         targetAAdress,
				Pid:             pndID,
				PluginId:        pluginID,
				MneName:         mneName,
				TransportOption: opt,
			},
		},
		Pid: pndID,
	}

	getRootPathRequest := &mnepb.GetPathRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Path:      "/",
		Pid:       pndID,
	}

	pathToChange, err := ygot.StringToStructuredPath("/")
	if err != nil {
		t.Error(err)
	}

	setPathListRequest := &mnepb.SetPathListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		ChangeRequest: []*mnepb.ChangeRequest{
			{
				Path:  pathToChange,
				Value: &gnmi.TypedValue{},
				ApiOp: operationUpdate,
			},
		},
	}

	// setup gRPC services
	mneService := mnepb.NewNetworkElementServiceClient(conn)

	// add one device to the controller
	addListResponse, err := mneService.AddList(ctx, addListRequest)
	if err != nil {
		t.Error(err)
	}

	// get root path
	getRootPathRequest.Mneid = addListResponse.GetResponses()[0].GetId()
	getPathResponse, err := mneService.GetPath(ctx, getRootPathRequest)
	if err != nil {
		t.Error(err)
	}

	setPathListRequest.ChangeRequest[0].Mneid = addListResponse.GetResponses()[0].GetId()

	// TODO: JsonIetfValue needs fixing
	setPathListRequest.ChangeRequest[0].Value = &gnmi.TypedValue{
		Value: &gnmi.TypedValue_JsonIetfVal{
			JsonIetfVal: []byte(getPathResponse.GetMneNotification()[0].GetUpdate()[0].GetVal().String()),
		},
	}

	_, err = mneService.SetPathList(ctx, setPathListRequest)
	// at the moment test is supposed to be successful if there is an error
	assert.NotEqual(t, nil, err)
}

func TestGetPathAndGetIntendedPathWithExternalChange(t *testing.T) {
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	// setup required parameters
	const initialName = "testTarget"

	opt := &tpb.TransportOption{
		Address:  targetAAdress,
		Username: targetUserAndPW,
		Password: targetUserAndPW,
		TransportOption: &tpb.TransportOption_GnmiTransportOption{
			GnmiTransportOption: &tpb.GnmiTransportOption{},
		},
		Tls: true,
	}

	addListRequest := &mnepb.AddListRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Mne: []*mnepb.SetMne{
			{
				Address:         targetAAdress,
				Pid:             pndID,
				PluginId:        pluginID,
				MneName:         initialName,
				TransportOption: opt,
			},
		},
		Pid: pndID,
	}

	getPathRequest := &mnepb.GetPathRequest{
		Timestamp: integration_test_utils.GetTimestamp(),
		Path:      "system/config/hostname",
		Pid:       pndID,
	}

	getIntendedPathRequest := &mnepb.GetIntendedPathRequest{
		Timestamp:    integration_test_utils.GetTimestamp(),
		IntendedPath: "system/config/hostname",
		Pid:          pndID,
	}

	// setup gRPC services
	mneService := mnepb.NewNetworkElementServiceClient(conn)

	// add device
	addResponse, err := mneService.AddList(ctx, addListRequest)
	if err != nil {
		t.Error(err)
		t.FailNow()
	}

	mneID := addResponse.GetResponses()[0].GetId()
	getPathRequest.Mneid = mneID
	getIntendedPathRequest.Mneid = mneID

	// get path and get intended path should be equal
	getPathResponse, err := mneService.GetPath(ctx, getPathRequest)
	if err != nil {
		t.Error(err)
		t.FailNow()
	}

	getIntendedPathResponse, err := mneService.GetIntendedPath(ctx, getIntendedPathRequest)
	if err != nil {
		t.Error(err)
		t.FailNow()
	}

	assert.Equal(t, getPathResponse.GetMneNotification()[0].GetUpdate()[0].GetVal().GetValue(), getIntendedPathResponse.GetMneNotification()[0].GetUpdate()[0].GetVal().GetValue())

	// change hostname via gnmic
	args := []string{"-a", targetA,
		"-u", "admin",
		"-p", "admin",
		"--tls-ca", "../../artifacts/ssl/gnmi-target/ca.crt",
		"--tls-cert", "../../artifacts/ssl/gnmi-target/certs/gnmi-target-selfsigned.crt",
		"--tls-key", "../../artifacts/ssl/gnmi-target/private/gnmi-target-selfsigned.key",
		"-e", "JSON_IETF",
		"set", "--update", "/system/config/hostname:::string:::diffHostname",
		"-d",
		"--skip-verify"}
	cmd := exec.Command("gnmic", args...)
	// prints cmd output to stdout for debugging purposes, comment/remove in case this clutters the output
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr

	err = cmd.Run()

	if err != nil {
		t.Error(err)
		t.FailNow()
	}

	// get path and get intended path should be different now
	getPathResponse, err = mneService.GetPath(ctx, getPathRequest)
	if err != nil {
		t.Error(err)
		t.FailNow()
	}

	getIntendedPathResponse, err = mneService.GetIntendedPath(ctx, getIntendedPathRequest)
	if err != nil {
		t.Error(err)
		t.FailNow()
	}

	assert.NotEqual(t, getPathResponse.GetMneNotification()[0].GetUpdate()[0].GetVal().GetValue(), getIntendedPathResponse.GetMneNotification()[0].GetUpdate()[0].GetVal().GetValue())
}
