package integration_test_topology

import (
	"context"
	"fmt"
	"testing"

	topologyPb "code.fbi.h-da.de/danet/gosdn/api/go/gosdn/topology"
	integration_test_utils "code.fbi.h-da.de/danet/gosdn/integration-tests/integrationTestUtils"
	"github.com/stretchr/testify/assert"
	"google.golang.org/grpc"
)

// The connection to the controller to use in each test.
var conn *grpc.ClientConn

// The context containing the credentials when authenticated.
var ctx context.Context

// A defaultSDN config with default/empty values.
var defaultSDNConfig string

func TestMain(m *testing.M) {
	localConn, localCtx, err := integration_test_utils.CreateSecureConnection()
	//localConn, localCtx, err := integration_test_utils.CreateConnection()
	if err != nil {
		fmt.Println(err.Error())
	}
	conn = localConn
	ctx = localCtx

	sndConfig, err := integration_test_utils.ExportCurrentSDNConfig(conn, ctx)
	defaultSDNConfig = sndConfig
	if err != nil {
		fmt.Println(err.Error())
	}
	m.Run()
}

func TestCreateTopologyAndGetIt(t *testing.T) {
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	addLinkRequest := topologyPb.AddLinkRequest{
		Link: &topologyPb.Link{
			Name:       "Link1",
			SourceNode: &topologyPb.Node{Name: "Node1"},
			TargetNode: &topologyPb.Node{Name: "Node2"},
			SourcePort: &topologyPb.Port{Name: "Port1",
				Configuration: &topologyPb.Configuration{
					Ip:           "10.0.0.1",
					PrefixLength: 24}},
			TargetPort: &topologyPb.Port{Name: "Port2",
				Configuration: &topologyPb.Configuration{
					Ip:           "10.0.0.2",
					PrefixLength: 24}}},
	}

	topologyService := topologyPb.NewTopologyServiceClient(conn)
	_, err := topologyService.AddLink(ctx, &addLinkRequest)
	if err != nil {
		t.Errorf("Error while adding link: %v", err)
	}

	getTopolgyResponse, err := topologyService.GetTopology(ctx, &topologyPb.GetTopologyRequest{})
	if err != nil {
		t.Errorf("Error while getting topology: %v", err)
	}

	assert.Equal(t, 1, len(getTopolgyResponse.GetToplogy().GetLinks()))
	assert.Equal(t, "Link1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetName())
	assert.Equal(t, "Node1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourceNode().GetName())
	assert.Equal(t, "Node2", getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetNode().GetName())
	assert.Equal(t, "Port1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourcePort().GetName())
	assert.Equal(t, "10.0.0.1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourcePort().GetConfiguration().GetIp())
	assert.Equal(t, int64(24), getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourcePort().GetConfiguration().GetPrefixLength())
	assert.Equal(t, "Port2", getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetPort().GetName())
	assert.Equal(t, "10.0.0.2", getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetPort().GetConfiguration().GetIp())
	assert.Equal(t, int64(24), getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetPort().GetConfiguration().GetPrefixLength())
}

func TestCreateTopologyFailsValidation(t *testing.T) {
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	addLinkRequest := topologyPb.AddLinkRequest{
		Link: &topologyPb.Link{
			Name:       "Link1",
			SourceNode: &topologyPb.Node{Name: "Node1"},
			TargetNode: &topologyPb.Node{Name: "Node2"},
			SourcePort: &topologyPb.Port{Name: "Port1"},
			TargetPort: &topologyPb.Port{Name: "Port2"}},
	}

	topologyService := topologyPb.NewTopologyServiceClient(conn)
	resp, err := topologyService.AddLink(ctx, &addLinkRequest)
	assert.Error(t, err, "Test expected an error, none received.")
	assert.Nil(t, resp, "Test expects the resp to be nil if an error is thrown beforehand.")
}

// Test that creates a topology and then deletes it and checks that all topologies are deleted.

// Can't delete topology because of bug in goSDN. If bug is fixed, uncomment and finish this test.
func TestCreateThenDeleteTopologyAndGetEmptyTopologies(t *testing.T) {
	defer integration_test_utils.ApplySDNConfig(conn, ctx, defaultSDNConfig)

	addLinkRequest := topologyPb.AddLinkRequest{
		Link: &topologyPb.Link{
			Id:         "8a01b8b0-6244-42bf-b72a-3821efc1a353",
			Name:       "Link1",
			SourceNode: &topologyPb.Node{Name: "Node1"},
			TargetNode: &topologyPb.Node{Name: "Node2"},
			SourcePort: &topologyPb.Port{Name: "Port1",
				Configuration: &topologyPb.Configuration{
					Ip:           "10.0.0.1",
					PrefixLength: 24}},
			TargetPort: &topologyPb.Port{Name: "Port2",
				Configuration: &topologyPb.Configuration{
					Ip:           "10.0.0.2",
					PrefixLength: 24}}},
	}

	topologyService := topologyPb.NewTopologyServiceClient(conn)
	_, err := topologyService.AddLink(ctx, &addLinkRequest)
	if err != nil {
		t.Errorf("Error while adding link: %v", err)
	}

	getTopolgyResponse, err := topologyService.GetTopology(ctx, &topologyPb.GetTopologyRequest{})
	if err != nil {
		t.Errorf("Error while getting topology: %v", err)
	}

	assert.Equal(t, 1, len(getTopolgyResponse.GetToplogy().GetLinks()))
	assert.Equal(t, "Link1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetName())
	assert.Equal(t, "Node1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourceNode().GetName())
	assert.Equal(t, "Node2", getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetNode().GetName())
	assert.Equal(t, "Port1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourcePort().GetName())
	assert.Equal(t, "10.0.0.1", getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourcePort().GetConfiguration().GetIp())
	assert.Equal(t, int64(24), getTopolgyResponse.GetToplogy().GetLinks()[0].GetSourcePort().GetConfiguration().GetPrefixLength())
	assert.Equal(t, "Port2", getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetPort().GetName())
	assert.Equal(t, "10.0.0.2", getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetPort().GetConfiguration().GetIp())
	assert.Equal(t, int64(24), getTopolgyResponse.GetToplogy().GetLinks()[0].GetTargetPort().GetConfiguration().GetPrefixLength())

	deleteLinkRequest := topologyPb.DeleteLinkRequest{Id: getTopolgyResponse.GetToplogy().GetLinks()[0].GetId()}
	_, err = topologyService.DeleteLink(ctx, &deleteLinkRequest)
	if err != nil {
		t.Errorf("Error while deleting link: %v", err)
	}
	getTopologyRequest := &topologyPb.GetTopologyRequest{}
	getTopologyResponse, err := topologyService.GetTopology(ctx, getTopologyRequest)
	if err != nil {
		t.Errorf("Error while getting topologies: %v", err)
	}

	assert.Equal(t, 0, len(getTopologyResponse.GetToplogy().GetLinks()))
}
